#lang racket/base

(require racket/match)

;;;;; AST

(struct Nil ())
(struct Pair (a b))
(struct Num (n))
(struct Data (l))
(struct First (p))
(struct Second (p))
(struct Let (n v e))
(struct Var (n))

;;;;; MIPS

(struct Move (rd rs))
(struct Li (r i))
(struct La (r a))
(struct Addi (rd rs i))
(struct Sw (r loc))
(struct Lw (r loc))
(struct Syscall ())
(struct Jr (r))
(struct Label (l))

;; addresses
(struct Lbl (l))   ;; label (souvent présent dans .data)
(struct Mem (b r)) ;; emplacement mémoire à l'adresse b + valeur du registre r

;;;;; minicomp
;; la convention utilisée dans ce compilateur est
;; de toujours mettre la valeur calculée dans $v0

(define (comp ast env fp-sp) ;; le décalage entre sp et fp est fp - sp
  (match ast
    ((Nil)
     ;; On représente Nil par l'adresse 0 (comme NULL en C)
     (comp (Num 0) env fp-sp))

    ((Num n)
     ;; Constante entière mise dans v0
     (list (Li 'v0 n)))

    ((Data d)
     ;; Pointeur dans .data mis dans v0
     (list (La 'v0 (Lbl d))))

    ((Pair a b)
     ;; Paire (a . b)
     (append
      ;; d'abord on compile a pour avoir sa valeur dans v0 :
      (comp a env fp-sp)
      ;; ensuite on l'empile :
      (list (Addi 'sp 'sp -4)
            (Sw 'v0 (Mem 0 'sp)))
      ;; ensuite on compile b pour l'avoir dans v0 en
      ;; se rappelant qu'on est décalé par rapport à fp :
      (comp b env (- fp-sp 4))
      ;; ensuite on compile la paire :
      (list (Lw 't0 (Mem 0 'sp)) ;; on dépile a
            (Addi 'sp 'sp 4)
            (Move 't1 'v0) ;; on récupère b
            (Li 'a0 8) ;; on alloue 8 octets = 2 mots mémoire
            (Li 'v0 9)
            (Syscall)
            ;; l'adresse de la paire est déjà dans v0
            (Sw 't0 (Mem 0 'v0)) ;; dans le premier on écrit a
            (Sw 't1 (Mem 4 'v0))))) ;; dans le second on écrit b

    ((First p)
     ;; Premier élément de la paire p
     (append
      ;; on compile p pour avoir l'adresse de la paire dans v0 :
      (comp p env fp-sp)
      ;; on récupère le premier dans v0 :
      (list (Lw 'v0 (Mem 0 'v0)))))

    ((Second p)
     ;; Second élément de la paire p
     (append
      ;; on compile p pour avoir l'adresse de la paire dans v0 :
      (comp p env fp-sp)
      ;; on récupère le second élément dans v0
      (list (Lw 'v0 (Mem 4 'v0)))))

    ((Let n v e)
     ;; Variable locale : let n = v in e
     (append
      ;; on compile v pour avoir sa valeur dans v0 :
      (comp v env fp-sp)
      ;; on empile la variable locale :
      (list (Addi 'sp 'sp -4)
            (Sw 'v0 (Mem 0 'sp)))
      ;; à partir de là fp - sp a grandi de 4 :
      (let ((fp-sp (- fp-sp 4)))
        ;; on compile e pour le mettre dans v0 :
        (comp e
              ;; en associant n à son adresse dans la pile par rapport
              ;; à fp, pour que cette adresse soit fixe
              (hash-set env n (Mem fp-sp 'fp))
              fp-sp))))

    ((Var n)
     ;; Référence à une variable
     ;; on met la valeur de la variable dans v0 :
     (list (Lw 'v0 (hash-ref env n))))))



(define (mips-loc loc)
  (match loc
    ((Lbl l)   (format "~a" l))
    ((Mem b r) (format "~a($~a)" b r))))

(define (mips-emit instr)
  (match instr
    ((Move rd rs)   (printf "move $~a, $~a\n" rd rs))
    ((Li r i)       (printf "li $~a, ~a\n" r i))
    ((La r a)       (printf "la $~a, ~a\n" r (mips-loc a)))
    ((Addi rd rs i) (printf "addi $~a, $~a, ~a\n" rd rs i))
    ((Sw r loc)     (printf "sw $~a, ~a\n" r (mips-loc loc)))
    ((Lw r loc)     (printf "lw $~a, ~a\n" r (mips-loc loc)))
    ((Syscall)      (printf "syscall\n"))
    ((Jr r)         (printf "jr $~a\n" r))
    ((Label l)      (printf "\t~a:\n" l))))

(define (mips-data data)
  (printf ".data\n")
  (hash-for-each data
                 (lambda (k v)
                   (printf "~a: .asciiz ~s\n" k v)))
  (printf "\n.text\n.globl main\nmain:\n"))


(mips-data (make-hash '((str_123 . "coucou") (nl . "\n"))))
(for-each mips-emit
          (append
           ;; On initialise notre environnement local :
           (list (Move 'fp 'sp))
           ;; On compile une expression :
           (comp (Let 'truc (Pair (Num 42) (Pair (Data 'str_123) (Nil)))
                      (Let 'bidule (Num 13)
                           (First (Second (Second (Pair (Var 'bidule) (Var 'truc)))))))
                 ;; avec un environnement vide :
                 (make-immutable-hash)
                 ;; et fp-sp = 0 (vu que fp = sp à ce moment là) :
                 0)
           ;; On affiche le résultat, qui est dans v0
           (list (Move 'a0 'v0)
                 (Li 'v0 4) ;; 4 pour print_string qui est le type du résultat
                 (Syscall)
                 ;; affichage retour à la ligne :
                 (La 'a0 (Lbl 'nl))
                 (Syscall)
                 ;; main return 0
                 (Li 'v0 0)
                 (Jr 'ra))))
